import gc #garbage collection, RAM management
gc.collect()
start_memory_free = gc.mem_free()
print("start memory free {} B".format( start_memory_free ))

# load system libraries
import os
import time
import board
import microcontroller
import digitalio
from analogio import AnalogIn
import neopixel
import sdioio
import storage
import busio
import displayio
displayio.release_displays()
import terminalio
from adafruit_display_text import label

# load device libraries
import adafruit_pcf8523
import adafruit_sht4x
import adafruit_tca9548a
from adafruit_as726x import AS726x_I2C
from adafruit_neokey.neokey1x4 import NeoKey1x4
from adafruit_ht16k33 import segments
#from adafruit_dps310.basic import DPS310 # barometer optional
#import adafruit_ssd1327 # OLED display. Initializing display disrupts sd card operation
#print("memory free after load libraries {} kB".format( gc.mem_free()/1000 ))

VIS_BANDS = ( 450, 500, 550, 570, 600, 650 ) # from amd as7262 datasheet
VIS_BAND_PREFIXES = ( "V", "B",  "G", "Y", "O", "R" )
NIR_BANDS = ( 610, 680, 730, 760, 810, 860 ) # from amd as7263 datasheet
ON = True       #   for active high discrete devices
OFF = False     #   for active high discrete devices
RED = ( 255, 0, 0 )
NEO_OFF = ( 0, 0, 0 )
DATA_FILE = "/sd/data.csv"

RECORD = const( 0 )
HOLD = const( 1 )
LIVE = const( 2 )
AVERAGE = const( 3 )

SHOW_UID = const( 0 )
SHOW_TIME = const( 1 )
SHOW_BATCH_NUMBER = const( 2 )

ORANGE = 0xFF5500
YELLOW = 0xFFFF00
GREEN = 0x00FF00
BLUE = 0x0000FF

def main():
    desired_sample_interval_s = 0.5
    averaging_count_down = 3
    count_down_step_s = 0.75
    averaging_number_of_samples = 20
    operational = True
    # initialize bus
    i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )
    if not i2c_bus:
        operational = False

    if operational:
        # initialize board devices
        UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
        print("unique identifier (UID) == {0}".format( UID ))
        power_indicator_LED = initialize_discrete_LED( board.LED )
        #print( "power_indicator_LED == {}".format( power_indicator_LED ))
        power_indicator_LED.value = ON
        main_battery_monitor = initialize_AnalogIn( board.VOLTAGE_MONITOR )
        main_battery_voltage = check_battery_voltage( main_battery_monitor )
        #print( "main_battery_voltage == {}".format( main_battery_voltage ))
        onboard_neopixel = initialize_neopixel( board.NEOPIXEL, 1 )
        set_neopixel( onboard_neopixel, RED )
        time.sleep( 0.1 )
        set_neopixel( onboard_neopixel, OFF )

    if operational:
        # initialize devices
        rtc_group = initialize_real_time_clock( i2c_bus )
        rtc = rtc_group[ 0 ]
        rtc_battery_OK = rtc_group[ 1 ]
        days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        # wday =    0         1         2           3            4          5          6
        four_digit_display = initialize_four_digit_display( i2c_bus )
        four_digit_display.print( UID )
        neokey1x4 = initialize_neokey1x4( i2c_bus )
        if neokey1x4:
            neokey1x4.pixels[0] = ORANGE
            neokey1x4.pixels[1] = YELLOW
            neokey1x4.pixels[2] = GREEN
            neokey1x4.pixels[3] = BLUE

    instrument_mode = LIVE
    last_instrument_mode = HOLD
    batch_number = 0

    display_mode = SHOW_UID
    number_of_display_modes = 3
    last_display_mode = 10 # startup
    display_mode_key_pressed = False
    last_display_mode_key_pressed = False
    toggle_record_pause_key_pressed = False
    last_toggle_record_pause_key_pressed = False
    sample_one_key_pressed = False
    last_sample_one_key_pressed = False
    sample_and_average_key_pressed = False
    last_sample_and_average_key_pressed = False



    while operational:
        if rtc:
            last_sample_time_s = time.monotonic()
            timestamp = rtc.datetime
            #print( "%s %d-%d-%d" % ( days[ timestamp.tm_wday ], timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday ))
            #print( "%d:%02d:%02d" % ( timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec ))
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
        iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
        weekday = timestamp_to_weekday( timestamp )
        decimal_hour = timestamp_to_decimal_hour( timestamp )
        if display_mode == SHOW_UID and last_display_mode != SHOW_UID:
            four_digit_display.fill(0)
            four_digit_display.print( UID )
            last_display_mode = SHOW_UID
        elif display_mode == SHOW_TIME:
            four_digit_display.fill(0)
            four_digit_display.print( "{:02}:{:02}".format( timestamp.tm_hour, timestamp.tm_min ))
            last_display_mode = SHOW_TIME
        elif display_mode == SHOW_BATCH_NUMBER:
            four_digit_display.fill(0)
            four_digit_display.print( batch_number )
            four_digit_display[0] = ( "B" )
            last_display_mode = SHOW_BATCH_NUMBER



        exit_wait = 0

        if instrument_mode == RECORD:
            print( "instrument mode == RECORD" )
            if last_instrument_mode != RECORD:
                print( "increment batch number ================================" )
                batch_number = batch_number + 1
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
            print( "Read data from sensors" )
            print( "Mirror data over USB" )
            print( "Write datapoint to file" )
            set_neopixel( onboard_neopixel, RED )
            time.sleep( 0.1 )
            set_neopixel( onboard_neopixel, OFF )
            last_instrument_mode = RECORD

        if instrument_mode == HOLD:
            print( "instrument mode == HOLD" )
            if last_instrument_mode != HOLD:
                print( "increment batch number ================================" )
                batch_number = batch_number + 1
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
                print( "Read data from sensors" )
                print( "Write datapoint to file" )
                set_neopixel( onboard_neopixel, RED )
                time.sleep( 0.1 )
                set_neopixel( onboard_neopixel, OFF )
            print( "Mirror data over USB" )
            last_instrument_mode = HOLD

        if instrument_mode == LIVE:
            print( "instrument mode == LIVE" )
            print( "Read data from sensors" )
            print( "Mirror data over USB" )
            last_instrument_mode = LIVE

        if instrument_mode == AVERAGE:
            four_digit_display.fill( 0 )
            print( "instrument mode == AVERAGE" )
            print( "increment batch number ================================" )
            batch_number = batch_number + 1
            exit_wait = 1
            count_down_index = averaging_count_down
            while count_down_index > 0:
                print( "show count down remaining" )
                four_digit_display.fill( 0 )
                four_digit_display[1] = "C"
                four_digit_display[2] ="D"
                four_digit_display.print( count_down_index )
                count_down_index = count_down_index - 1
                time.sleep( count_down_step_s )
            sampling_index = averaging_number_of_samples
            while sampling_index > 0:
                print( "sampling index == {}".format( sampling_index ))
                print( "Read data from sensors" )
                print( "Write datapoint to file" )
                four_digit_display.fill( 0 )
                four_digit_display.print( sampling_index )
                set_neopixel( onboard_neopixel, RED )
                time.sleep( 0.1 )
                set_neopixel( onboard_neopixel, OFF )
                sampling_index = sampling_index - 1
            print( "record average values and standard deviations too" )
            four_digit_display.print( "DONE" )
            time.sleep( 1 )
            display_mode = SHOW_BATCH_NUMBER
            last_instrument_mode = HOLD
            instrument_mode = HOLD

        # Wait interval
        remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
        while remaining_wait_time > 0 and exit_wait == 0:

            # check for keys pressed
            display_mode_key_pressed, sample_one_key_pressed, sample_and_average_key_pressed, toggle_record_pause_key_pressed = check_neokey( neokey1x4 )
            # turn key light off to acknowledge press detected
            if display_mode_key_pressed:
                neokey1x4.pixels[0] = 0x000000
            else:
                neokey1x4.pixels[0] = 0xFF5500
            if sample_one_key_pressed:
                neokey1x4.pixels[1] = 0x000000
            else:
                neokey1x4.pixels[1] = 0xFFFF00
            if sample_and_average_key_pressed:
                neokey1x4.pixels[2] = 0x000000
            else:
                neokey1x4.pixels[2] = 0x00FF00
            if toggle_record_pause_key_pressed:
                neokey1x4.pixels[3] = 0x000000
            else:
                neokey1x4.pixels[3] = 0x0000FF


            if display_mode_key_pressed and not last_display_mode_key_pressed:
                display_mode = ( display_mode + 1 ) % number_of_display_modes
                if instrument_mode == HOLD:
                    instrument_mode = LIVE
            last_display_mode_key_pressed = display_mode_key_pressed

            if sample_one_key_pressed and not last_sample_one_key_pressed:
                if instrument_mode != HOLD:
                    instrument_mode = HOLD
                last_instrument_mode = RECORD
            last_sample_one_key_pressed = sample_one_key_pressed

            if toggle_record_pause_key_pressed and not last_toggle_record_pause_key_pressed:
                if instrument_mode != RECORD:
                    instrument_mode = RECORD
                else:
                    instrument_mode = LIVE
            last_toggle_record_pause_key_pressed = toggle_record_pause_key_pressed

            if sample_and_average_key_pressed and not last_sample_and_average_key_pressed:
                if instrument_mode != AVERAGE:
                    instrument_mode = AVERAGE
            last_sample_and_average_key_pressed = sample_and_average_key_pressed

            #print( display_mode )
            remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
            #print( "remaining_wait_time == {:}".format(remaining_wait_time))
            time.sleep( 0.1 )

    print( "STELLA done" )

# function definitions
def check_neokey( neokey_object ):
    keys = [ False, False, False, False ]
    try:
        keys[ 0 ] = neokey_object[ 0 ]
        keys[ 1 ] = neokey_object[ 1 ]
        keys[ 2 ] = neokey_object[ 2 ]
        keys[ 3 ] = neokey_object[ 3 ]
    except ValueError:
        keys = [ False, False, False, False ]
    return keys


def initialize_four_digit_display( i2c_bus ):
    try:
        display = segments.Seg7x4( i2c_bus )
        display.fill(0)
    except ValueError:
        print( "4 digit 7 segment display not found" )
        display = False
    return display

def initialize_neokey1x4( i2c_bus ):
    try:
        neokey = NeoKey1x4(i2c_bus, addr=0x30)
    except ValueError:
        print( "neokey 1x4 module not found" )
        neokey = False
    return neokey

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def initialize_real_time_clock( i2c_bus ):
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        clock_battery_OK = not rtc.battery_low
        if clock_battery_OK:
            print( "clock battery is OK." )
        else:
            print( "clock battery is low. Replace the clock battery." )
    except NameError:
        print( "real time clock failed to initialize" )
        rtc = False
    return ( rtc, clock_battery_OK )

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
    except ValueError as err:
        print( "i2c bus failed to initialize '{}'".format( err ))
        i2c_bus = False
    return i2c_bus


def set_neopixel( neopixel_object, color ):
    if neopixel_object:
        if color:
            neopixel_object[ 0 ] = color
        else:
            neopixel_object[ 0 ] = ( 0, 0, 0 )
        neopixel_object.show()
    else:
        return False

def initialize_neopixel( pin_number, number_of_pixels ):
    try:
        neopixel_object = neopixel.NeoPixel( pin_number, number_of_pixels, brightness=1, auto_write=False, pixel_order=neopixel.GRB )
    except ValueError:
        print( "neopixel initialization failed." )
        neopixel_object = False
    return neopixel_object

def check_battery_voltage( battery_monitor ):
    if battery_monitor:
        battery_voltage = round(( battery_monitor.value * 3.3) / 65536 * 2, 2 )
        return battery_voltage
    else:
        return 0

def initialize_AnalogIn( pin_number ):
    try:
        analog_signal = AnalogIn( pin_number )
    except ValueError:
        print( "analog input initialization failed." )
        analog_signal = False
    return analog_signal

def initialize_discrete_LED( pin_number ):
    try:
        discrete_LED = digitalio.DigitalInOut( pin_number )
        discrete_LED.direction = digitalio.Direction.OUTPUT
    except ValueError:
        print( "discrete LED initialization failed." )
        discrete_LED = False
    return discrete_LED

def stall():
    print("intentionally stalled, press return to continue")
    input_string = False
    while input_string == False:
        input_string = input().strip()

main()
