# STELLA-Q software, version 1.0
# Science and Technology Education for Land/ Life Assessment
# Paul Mirel 2023-01-08

import gc #garbage collection, RAM management
gc.collect()
start_memory_free = gc.mem_free()
print("start memory free {} B".format( start_memory_free ))

# load system libraries
import os
import time
import board
import microcontroller
import digitalio
from analogio import AnalogIn
import neopixel
import sdioio
import storage
import busio
import displayio
displayio.release_displays()
import terminalio
from adafruit_display_text import label

# load device libraries
import adafruit_pcf8523
import adafruit_sht4x
import adafruit_tca9548a
from adafruit_as726x import AS726x_I2C
from adafruit_neokey.neokey1x4 import NeoKey1x4
from adafruit_ht16k33 import segments
#from adafruit_dps310.basic import DPS310 # barometer optional
#import adafruit_ssd1327 # OLED display. Initializing display disrupts sd card operation
#print("memory free after load libraries {} kB".format( gc.mem_free()/1000 ))

VIS_BANDS = ( 450, 500, 550, 570, 600, 650 ) # from amd as7262 datasheet
VIS_BAND_PREFIXES = ( "V", "B",  "G", "Y", "O", "R" )
NIR_BANDS = ( 610, 680, 730, 760, 810, 860 ) # from amd as7263 datasheet
ON = True       #   for active high discrete devices
OFF = False     #   for active high discrete devices
RED = ( 255, 0, 0 )
NEO_OFF = ( 0, 0, 0 )
DATA_FILE = "/sd/data.csv"

RECORD = const( 0 )
HOLD = const( 1 )
LIVE = const( 2 )
AVERAGE = const( 3 )

SHOW_UID = const( 0 )
SHOW_TIME = const( 1 )
SHOW_BATCH_NUMBER = const( 2 )

ORANGE = 0xFF5500
YELLOW = 0xFFFF00
GREEN = 0x00FF00
BLUE = 0x0000FF

mux_channel_NIR = const( 0 )
mux_channel_VIS = const( 1 )

def main():
    operational = True
    # initialize bus
    i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )
    if not i2c_bus:
        operational = False

    if operational:
        # initialize board devices
        UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
        print("unique identifier (UID) == {0}".format( UID ))
        power_indicator_LED = initialize_discrete_LED( board.LED )
        #print( "power_indicator_LED == {}".format( power_indicator_LED ))
        power_indicator_LED.value = ON
        main_battery_monitor = initialize_AnalogIn( board.VOLTAGE_MONITOR )
        main_battery_voltage = check_battery_voltage( main_battery_monitor )
        #print( "main_battery_voltage == {}".format( main_battery_voltage ))
        onboard_neopixel = initialize_neopixel( board.NEOPIXEL, 1 )
        set_neopixel( onboard_neopixel, RED )
        time.sleep( 0.1 )
        set_neopixel( onboard_neopixel, OFF )

    if operational:
        # initialize devices
        rtc_group = initialize_real_time_clock( i2c_bus )
        rtc = rtc_group[ 0 ]
        rtc_battery_OK = rtc_group[ 1 ]
        days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        # wday =    0         1         2           3            4          5          6

        # initialize sd card file storage
        vfs = initialize_virtual_file_system()
        if vfs:
            storage.mount(vfs, "/sd")
        else:
            set_neopixel( onboard_neopixel, RED )
            print( "(future print to instrument display): SD card initialization failed. No SD card or card full." )
        write_header = initialize_data_file()
        print( "write header == {}".format( write_header ))

        if rtc:
            timestamp = rtc.datetime
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
        batch_number = update_batch( timestamp )

        # gather STELLA instrument config
        desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s = check_STELLA_config()


        # initialize sensors
        temp_hum_SHT40 = initialize_SHT40( i2c_bus )
        #baro_alt_DPS310 = initialize_DPS310( i2c_bus )
        bus_mux = initialize_tca_i2c_multiplexer( i2c_bus )
        NIR_spectral_sensor = initialize_spectral_sensor( bus_mux, mux_channel_NIR )
        VIS_spectral_sensor = initialize_spectral_sensor( bus_mux, mux_channel_VIS )

        # set instrument mode
        instrument_mode = LIVE
        last_instrument_mode = HOLD

        four_digit_display = initialize_four_digit_display( i2c_bus )
        if four_digit_display:
            four_digit_display.print( UID )
        neokey1x4 = initialize_neokey1x4( i2c_bus )
        if neokey1x4:
            neokey1x4.pixels[0] = ORANGE
            neokey1x4.pixels[1] = YELLOW
            neokey1x4.pixels[2] = GREEN
            neokey1x4.pixels[3] = BLUE

        # set display modes
        display_mode = SHOW_UID
        number_of_display_modes = 3
        last_display_mode = 10 # startup
        display_mode_key_pressed = False
        last_display_mode_key_pressed = False
        toggle_record_pause_key_pressed = False
        last_toggle_record_pause_key_pressed = False
        sample_one_key_pressed = False
        last_sample_one_key_pressed = False
        sample_and_average_key_pressed = False
        last_sample_and_average_key_pressed = False

    while operational:
        gc.collect()
        if rtc:
            last_sample_time_s = time.monotonic()
            timestamp = rtc.datetime
            #print( "%s %d-%d-%d" % ( days[ timestamp.tm_wday ], timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday ))
            #print( "%d:%02d:%02d" % ( timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec ))
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
        iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
        weekday = timestamp_to_weekday( timestamp )
        decimal_hour = timestamp_to_decimal_hour( timestamp )

        if four_digit_display:
            if display_mode == SHOW_UID and last_display_mode != SHOW_UID:
                four_digit_display.fill(0)
                four_digit_display.print( UID )
                last_display_mode = SHOW_UID
            elif display_mode == SHOW_TIME:
                four_digit_display.fill(0)
                four_digit_display.print( "{:02}:{:02}".format( timestamp.tm_hour, timestamp.tm_min ))
                last_display_mode = SHOW_TIME
            elif display_mode == SHOW_BATCH_NUMBER:
                four_digit_display.fill(0)
                four_digit_display.print( batch_number )
                four_digit_display[0] = ( "B" )
                last_display_mode = SHOW_BATCH_NUMBER

        exit_wait = 0

        if instrument_mode == RECORD:
            #print( "instrument mode == RECORD" )
            if last_instrument_mode != RECORD:
                batch_number = update_batch( timestamp )
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
            if VIS_spectral_sensor:
                visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
            else:
                visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            if NIR_spectral_sensor:
                nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
            else:
                nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
            main_battery_voltage = check_battery_voltage( main_battery_monitor )
            datapoint = build_datapoint( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
            write_line_success = write_line_to_file( onboard_neopixel, datapoint )
            mirror_data_over_usb( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
            last_instrument_mode = RECORD

        if instrument_mode == HOLD:
            #print( "instrument mode == HOLD" )
            if last_instrument_mode != HOLD:
                batch_number = update_batch( timestamp )
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
                if VIS_spectral_sensor:
                    visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                else:
                    visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                if NIR_spectral_sensor:
                    nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                else:
                    nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                datapoint = build_datapoint( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
                write_line_success = write_line_to_file( onboard_neopixel, datapoint )
                mirror_data_over_usb( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
            last_instrument_mode = HOLD

        if instrument_mode == LIVE:
            #print( "instrument mode == LIVE" )
            if VIS_spectral_sensor:
                visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
            else:
                visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            if NIR_spectral_sensor:
                nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
            else:
                nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
            main_battery_voltage = check_battery_voltage( main_battery_monitor )
            mirror_data_over_usb( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
            last_instrument_mode = LIVE

        if instrument_mode == AVERAGE and four_digit_display:
            if vfs:
                neokey1x4.pixels[2] = 0x00FF00
                four_digit_display.fill( 0 )
                #print( "instrument mode == AVERAGE" )
                batch_number = update_batch( timestamp )
                exit_wait = 1
                count_down_index = averaging_count_down
                temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                VIS_sensor_temperature, NIR_sensor_temperature = get_spectral_sensor_temperatures( VIS_spectral_sensor, NIR_spectral_sensor )
                header_values = ( UID, batch_number, iso8601_utc, temperature_C, relative_humidity_percent, VIS_sensor_temperature, NIR_sensor_temperature )
                sampling_filename = create_sampling_file( UID, timestamp, decimal_hour, header_values )
                while count_down_index > 0:
                    print( "count down remaining = {}".format( count_down_index ))
                    four_digit_display.fill( 0 )
                    four_digit_display[1] = "C"
                    four_digit_display[2] ="D"
                    four_digit_display.print( count_down_index )
                    count_down_index = count_down_index - 1
                    time.sleep( count_down_step_s )
                sampling_index = averaging_number_of_samples
                sample_average = [ 0,0,0,0,0,0,0,0,0,0,0,0 ]
                while sampling_index > 0:
                    print( "sample number = {}".format( sampling_index ))
                    if VIS_spectral_sensor:
                        visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                    else:
                        visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                    if NIR_spectral_sensor:
                        nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                    else:
                        nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                    write_sampling_line_to_file( onboard_neopixel, sampling_filename, sampling_index, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared )
                    spectral_samples = visible_spectrum_uW_per_cm_squared + nir_spectrum_uW_per_cm_squared
                    #print( spectral_samples )
                    for i,v in enumerate( spectral_samples ):
                        sample_average[ i ] = sample_average[ i ] + v
                    four_digit_display.fill( 0 )
                    four_digit_display.print( sampling_index )
                    sampling_index = sampling_index - 1
                for i,v in enumerate( sample_average ):
                    sample_average[ i ] = round( sample_average[ i ]/ averaging_number_of_samples, 1 )
                #print( sample_average )
                write_sampling_averages_to_file( onboard_neopixel, sampling_filename, averaging_number_of_samples, sample_average )
                four_digit_display.print( "DONE" )
                averaged_visible_spectrum_uW_per_cm_squared = [0,0,0,0,0,0]
                averaged_nir_spectrum_uW_per_cm_squared = [0,0,0,0,0,0]
                for i,v in enumerate( sample_average ):
                    if i < 6:
                        averaged_visible_spectrum_uW_per_cm_squared[ i ] = sample_average[ i ]
                    else:
                        averaged_nir_spectrum_uW_per_cm_squared[ i - 6 ] = sample_average[ i ]
                datapoint = build_datapoint( UID, timestamp, decimal_hour, batch_number, averaged_visible_spectrum_uW_per_cm_squared, averaged_nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
                write_line_success = write_line_to_file( onboard_neopixel, datapoint )
                gc.collect()
                mirror_data_over_usb( UID, timestamp, decimal_hour, batch_number, averaged_visible_spectrum_uW_per_cm_squared, averaged_nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage)
                time.sleep( 1 )
                display_mode = SHOW_BATCH_NUMBER
                last_instrument_mode = HOLD
                instrument_mode = HOLD
            else:
                print( "flash key error on key[2]" )
                print( "print 'no sd' to display" )

        # Wait loop
        gc.collect()
        wait_memory_free = gc.mem_free()
        #print("at wait loop, memory free {} B".format( wait_memory_free ))
        remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
        while remaining_wait_time > 0 and exit_wait == 0:

            # check for keys pressed
            display_mode_key_pressed, sample_one_key_pressed, sample_and_average_key_pressed, toggle_record_pause_key_pressed = check_neokey( neokey1x4 )
            # turn key light off to acknowledge press detected
            if display_mode_key_pressed:
                neokey1x4.pixels[0] = 0x000000
            else:
                neokey1x4.pixels[0] = 0xFF5500
            if sample_one_key_pressed:
                neokey1x4.pixels[1] = 0x000000
            else:
                neokey1x4.pixels[1] = 0xFFFF00
            if sample_and_average_key_pressed:
                neokey1x4.pixels[2] = 0x000000
            else:
                neokey1x4.pixels[2] = 0x00FF00
            if toggle_record_pause_key_pressed:
                neokey1x4.pixels[3] = 0x000000
            else:
                neokey1x4.pixels[3] = 0x0000FF


            if display_mode_key_pressed and not last_display_mode_key_pressed:
                display_mode = ( display_mode + 1 ) % number_of_display_modes
                if instrument_mode == HOLD:
                    instrument_mode = LIVE
            last_display_mode_key_pressed = display_mode_key_pressed

            if sample_one_key_pressed and not last_sample_one_key_pressed:
                if instrument_mode != HOLD:
                    instrument_mode = HOLD
                last_instrument_mode = RECORD
            last_sample_one_key_pressed = sample_one_key_pressed

            if toggle_record_pause_key_pressed and not last_toggle_record_pause_key_pressed:
                if instrument_mode != RECORD:
                    instrument_mode = RECORD
                else:
                    instrument_mode = LIVE
            last_toggle_record_pause_key_pressed = toggle_record_pause_key_pressed

            if sample_and_average_key_pressed and not last_sample_and_average_key_pressed:
                if instrument_mode != AVERAGE:
                    instrument_mode = AVERAGE
            last_sample_and_average_key_pressed = sample_and_average_key_pressed

            #print( display_mode )
            remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
            #print( "remaining_wait_time == {:}".format(remaining_wait_time))
            time.sleep( 0.1 )

    print( "STELLA done" )

# function definitions
def write_sampling_averages_to_file( onboard_neopixel, sampling_filename, averaging_number_of_samples, sample_average ):
    if onboard_neopixel:
        set_neopixel( onboard_neopixel, RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( averaging_number_of_samples ))
            sf.write( ", " )
            for i,v in enumerate( sample_average ):
                sf.write( str(v) )
                if i < len(sample_average)-1: # no trailing ','
                    sf.write(",")
            #sf.write( "\n" )
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def write_sampling_line_to_file( onboard_neopixel, sampling_filename, sampling_index, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared ):
    if onboard_neopixel:
        set_neopixel( onboard_neopixel, RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( sampling_index ))
            sf.write( ", " )
            for i,v in enumerate(visible_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                sf.write(",")
            for i,v in enumerate(nir_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                if i < len(nir_spectrum_uW_per_cm_squared)-1: # no trailing ','
                    sf.write(",")
            sf.write( "\n" )
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def get_spectral_sensor_temperatures(VIS_sensor, NIR_sensor):
    vis_temp_C = round( VIS_sensor.temperature, 1 )
    nir_temp_C = round( NIR_sensor.temperature, 1 )
    return vis_temp_C, nir_temp_C

def create_sampling_file( UID, timestamp, decimal_hour, values ):
    sampling_filename = (
        "/sd/" + str(UID) + "_" + str(timestamp.tm_year)
        + str("{:02d}".format( timestamp.tm_mon)) + str("{:02d}".format(timestamp.tm_mday))
        + "_" + str(decimal_hour) + ".csv" )
    try:
        with open( sampling_filename, "w" ) as sf:
            sf.write( "UID, batch_number, iso8601_utc, air_temperature, relative_humidity, VIS_sensor_temperature, NIR_sensor_temperature\n" )
            for i,v in enumerate(values):
                sf.write( str(v) )
                if i < len(values)-1: # no trailing ','
                    sf.write(",")
            sf.write( "\n" )
            sf.write( "sample_number, " )
            for i,v in enumerate(VIS_BANDS):
                sf.write( str(v) )
                sf.write(",")
            for i,v in enumerate(NIR_BANDS):
                #print(v)
                sf.write( str(v) )
                if i < len(values)-2: # no trailing ','
                    sf.write(",")
                    #print(",")
            sf.write( "\n" )
        return sampling_filename
    except OSError:
        return False

def initialize_data_file():
    need_header = False
    try:
        os.stat( DATA_FILE ) # fail if data.csv file is not already there
        #raise OSError # uncomment to force header everytime
        return False
    except OSError:
        # setup the header for first time
        need_header = True
        #print( "need header = True")
        if need_header:
            preheader_mem_free = gc.mem_free()
            print( "preheader mem free = {}".format(preheader_mem_free))
            gc.collect()
            try:
                with open( DATA_FILE, "w" ) as f:
                    header = (
                        "UID, batch_number, weekday, timestamp, hour_format, decimal_hour, "+
                        "air_temperature_name, air_temperature_units, air_temperature, air_temperature_calibration, "+
                        "relative_humidity_name, relative_humidity_units, relative_humidity, relative_humidity_calibration, "+
                        "visible_spectrum_name, visible_spectrum_wavelength, visible_spectrum_units, visible_spectrum_calibration, "+
                        "V450_wavelength, V450_irradiance, B500_wavelength, B500_irradiance, G550_wavelength, G550_irradiance, "+
                        "Y570_wavelength, Y570_irradiance, O600_wavelength, O600_irradiance, R650_wavelength, R650_irradiance, "+
                        "near_infrared_spectrum_name, near_infrared_spectrum_wavelength, near_infrared_spectrum_units, near_infrared_spectrum_calibration, "+
                        "nir610_wavelength, nir610_irradiance, nir680_wavelength, nir680_irradiance, nir730_wavelength, nir730_irradiance, "+
                        "nir760_wavelength, nir760_irradiance, nir810_wavelength, nir810_irradiance, nir860_wavelength, nir860_irradiance, "+
                        "battery_voltage_name, battery_voltage_units, battery_voltage"+
                        "\n"
                        )

                    f.write( header )
            except OSError:
                pass
            postheader_mem_free = gc.mem_free()
            print( "header mem usage = {}".format(postheader_mem_free - preheader_mem_free))
        return True


def write_line_to_file( onboard_neopixel, values ):
    gc.collect()
    try:
        # with sys.stdout as f: # uncomment to write to console (comment "with open")
        file_handling_last_sample_time_s = time.monotonic()
        with open( DATA_FILE, "a" ) as f: # open seems about TBD secs.
            last_sample_time_s = time.monotonic()
            if onboard_neopixel:
                set_neopixel( onboard_neopixel, RED )
            for i,v in enumerate(values): # 0.16 secs to write
                f.write( str(v) )
                #print( str(v) )
                if i < len(values)-1: # no trailing ','
                    f.write(",")
            f.write("\n")
        f.close()
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        write_command_time = time.monotonic() - last_sample_time_s
        #print ( "write command time s = {}".format(write_command_time))
        file_write_time = time.monotonic() - file_handling_last_sample_time_s
        print ( "File write time s = {}".format(file_write_time))
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "Error: sd card fail: {:} ".format(err) )
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, RED ) # constant ON to show error, likely no SD card present, or SD card full.
        return False
    return True

def build_datapoint( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage):
    iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
    weekday = timestamp_to_weekday( timestamp )
    decimal_hour = timestamp_to_decimal_hour( timestamp )
    datapoint = [] #reset datapoint string
    #datapoint.append( "UID:" )
    datapoint.append( UID )
    datapoint.append( batch_number )
    datapoint.append( "{:},{:},hh.hhhh,{:}".format(weekday, iso8601_utc, decimal_hour ) )
    # name,units,value,calibration_error
    #datapoint.append( "surface_temperature,C,{:},1.0".format( surface_temperature_C ) )
    datapoint.append( "air_temperature,C,{:},0.3".format( temperature_C ) )
    datapoint.append( "relative_humidity,%,{:},1.8".format(  relative_humidity_percent ) )
    # spectrum, wavelength-units, irradiance-units, bandlabel, irradiance, cal-error
    datapoint.append( "visible_spectrum,nm,uW/cm^2,12/100")
    # VIS_BANDS have a color initial prefix
    for i,band in enumerate( VIS_BANDS ):
        datapoint.append( band )
        datapoint.append( visible_spectrum_uW_per_cm_squared[ i ] )
    datapoint.append( "near_infrared_spectrum,nm,uW/cm^2,12/100")
    for i,band in enumerate( NIR_BANDS ):
        datapoint.append( band )
        datapoint.append( nir_spectrum_uW_per_cm_squared[ i ] )
    datapoint.append( "battery_voltage,V,{:}".format( main_battery_voltage ) )
    return datapoint

def mirror_data_over_usb( UID, timestamp, decimal_hour, batch_number, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, temperature_C, relative_humidity_percent, main_battery_voltage):
    mirror_data = "mirror_data, {}, {}".format("UID: " + str(UID), "batch: " + str(batch_number))
    mirror_data = ( mirror_data
        + ", " + "year: " + str( timestamp.tm_year )
        + ", " + "month: "+ str( timestamp.tm_mon )
        + ", " + "day: "+ str( timestamp.tm_mday )
        + ", " + "hour: "+ str( timestamp.tm_hour )
        + ", " + "min: "+ str( timestamp.tm_min )
        + ", " + "sec: "+ str( timestamp.tm_sec )
        + ", " + "dec_hour: "+ str( decimal_hour )
        + ", " + "v450: "+ str( visible_spectrum_uW_per_cm_squared[ 0 ] )
        + ", " + "b500: "+ str( visible_spectrum_uW_per_cm_squared[ 1 ] )
        + ", " + "g550: "+ str( visible_spectrum_uW_per_cm_squared[ 2 ] )
        + ", " + "y570: "+ str( visible_spectrum_uW_per_cm_squared[ 3 ] )
        + ", " + "o600: "+ str( visible_spectrum_uW_per_cm_squared[ 4 ] )
        + ", " + "r650: "+ str( visible_spectrum_uW_per_cm_squared[ 5 ] )
        + ", " + "610: "+ str( nir_spectrum_uW_per_cm_squared[ 0 ] )
        + ", " + "680: "+ str( nir_spectrum_uW_per_cm_squared[ 1 ] )
        + ", " + "730: "+ str( nir_spectrum_uW_per_cm_squared[ 2 ] )
        + ", " + "760: "+ str( nir_spectrum_uW_per_cm_squared[ 3 ] )
        + ", " + "810: "+ str( nir_spectrum_uW_per_cm_squared[ 4 ] )
        + ", " + "860: "+ str( nir_spectrum_uW_per_cm_squared[ 5 ] )
        #+ ", " + "surface_temp: " + str(surface_temperature_C)
        + ", " + "air_temp: " + str( temperature_C )
        + ", " + "rel_humidity: " + str( relative_humidity_percent )
        + ", " + "main_bat_volts: " + str( main_battery_voltage )
        )
    print( mirror_data )
    return True


def read_environmental_sensors( temp_hum_SHT40 ):
    if temp_hum_SHT40:
        temperature_C, relative_humidity_percent = temp_hum_SHT40.measurements
        temperature_C = round( temperature_C, 1 )
        relative_humidity_percent = round( relative_humidity_percent, 1 )
    else:
        temperature_C = 0
        relative_humidity_percent = 0
    return temperature_C, relative_humidity_percent

def read_spectrum( spectral_sensor ):
    decimal_places = 1
    # returns readings in a tuple, six channels of data
    # Calibration information from AS7262 datasheet:
    # Each channel is tested with GAIN = 16x, Integration Time (INT_T) = 166ms and VDD = VDD1 = VDD2 = 3.3V, TAMB=25°C.
    # The accuracy of the channel counts/μW/cm2 is ±12%.
    # Sensor Mode 2 is a continuous conversion of light into data on all 6 channels
    # 450nm, 500nm, 550nm, 570nm, 600nm and 650nm
    # sensor.violet returns the calibrated floating point value in the violet channel.
    # sensor.raw_violet returns the uncalibrated decimal count value in the violet channel.
    # that syntax is the same for each of the 6 channels

    # NOTE: the library was written for the visible sensor, as7262.
    # The library works exactly the same for the near infrared sensor, as7263,
    # but the band names are called as if the sensor were the visible one.
    try:
        while not spectral_sensor.data_ready:
            time.sleep(0.01)
        initial_read_time_s = time.monotonic()
        violet_calibrated = round( spectral_sensor.violet, decimal_places )
        blue_calibrated = round( spectral_sensor.blue, decimal_places )
        green_calibrated = round( spectral_sensor.green, decimal_places )
        yellow_calibrated = round( spectral_sensor.yellow, decimal_places )
        orange_calibrated = round( spectral_sensor.orange, decimal_places )
        red_calibrated = round( spectral_sensor.red, decimal_places )
        final_read_time_s = time.monotonic()
        read_time_s = final_read_time_s - initial_read_time_s
        #print( "Sensor read time s = {}".format( read_time_s ))
        spectral_tuple = ( violet_calibrated, blue_calibrated, green_calibrated, yellow_calibrated, orange_calibrated, red_calibrated )
        return spectral_tuple
    except ValueError as err:
        print( "Error: spectral sensor fail: {:}".format(err) )
        return ( 0, 0, 0, 0, 0, 0)

def initialize_spectral_sensor( bus_mux, mux_channel ):
    try:
        spectral_sensor = AS726x_I2C( bus_mux[ mux_channel ])
        print( "\nSet spectral sensor at mux channel {} to 6x gain and 166 ms integration time, because those are the values used at the factory during the sensor calibration.".format( mux_channel ))
        spectral_sensor.conversion_mode = spectral_sensor.MODE_2
        print( "Default Gain = {}".format( spectral_sensor.gain ))
        spectral_sensor.gain = 16
        print( "Gain Now = {}".format( spectral_sensor.gain ))
        print( "Default Integration Time = {} ms".format( spectral_sensor.integration_time ))
        spectral_sensor.integration_time = 166 # from 1 to 714 = 255 * 2.8ms ms
        print( "Integration Time Now = {} ms".format( spectral_sensor.integration_time ))
        return spectral_sensor
    except ValueError:
        print( "spectral_sensor at channel {} not found".format( mux_channel ))
        return False

def initialize_tca_i2c_multiplexer( i2c_bus ):
    try:
        tca = adafruit_tca9548a.TCA9548A( i2c_bus )
    except ValueError:
        print( "i2c bus multiplexer not available" )
        tca = False
    return tca

'''
def initialize_DPS310( i2c_bus ):
    try:
        dps310 = DPS310( i2c_bus )
    except RuntimeError:
        print( "dps310 barometer/ altimeter could not be found" )
        dps310 = False
    return dps310
'''

def initialize_SHT40( i2c_bus ):
    try:
        sht = adafruit_sht4x.SHT4x( i2c_bus )
    except ValueError:
        print( "sht40 temp_humid could not be found" )
        sht = False
    try:
        print( "Found SHT40 with serial number", hex( sht.serial_number ))
        sht.mode = adafruit_sht4x.Mode.NOHEAT_HIGHPRECISION
        # Can also set the mode to enable heater
        # sht.mode = adafruit_sht4x.Mode.LOWHEAT_100MS
        print("SHT40 mode is: ", adafruit_sht4x.Mode.string[ sht.mode ])
    except OSError:
        print( "SHT40 sensor serial number lookup failed" )
    return sht

def check_STELLA_config():
    try:
        with open( "/sd/STELLA_config.txt", "r" ) as s:
            try:
                header = s.readline()
                #print( sampling_header )
                values = s.readline()
                #print(sampling_values)
                values= values.rstrip('\n')
                values = values.split(',')
                #print(line)
                desired_sample_interval_s = float( values[ 0 ])
                averaging_number_of_samples = int( values[ 1 ])
                averaging_count_down = int( values[ 2 ])
                count_down_step_s = float( values[ 3 ])

            # TBD: catch error when /sd doesn't exist
            except ValueError:
                pass
    except OSError:
            print( "sampling_config.txt file not found" )
            try:
                with open( "/sd/STELLA_config.txt", "w" ) as s:
                    s.write( "desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s\n" )
                    s.write( "0.75, 20, 3, 0.75\n" )
                print("created new STELLA_config.txt file")
            except OSError:
                pass
            desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s = 0.75, 21, 3, 0.75
    return desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s

def initialize_virtual_file_system():
    try:
        sdcard = sdioio.SDCard( clock=board.SDIO_CLOCK, command=board.SDIO_COMMAND, data=board.SDIO_DATA, frequency=25000000 )
        vfs = storage.VfsFat(sdcard)
        #os.listdir('/sd')
        print( "sdcard file system frequency == {}".format( sdcard.frequency ))
        return vfs
    except OSError:
        print( "SD card failed initialization: OSError" )
        return False
    except ValueError:
        print( "SD card failed initialization: ValueError" )
        return False

def update_batch( timestamp ):
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            # TBD: catch error when /sd doesn't exist
            except ValueError:
                previous_batch_number = 0
                # corrupted data in batch number file, setting batch to 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0
    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
        # TBD: catch error when /sd doesn't exist
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def check_neokey( neokey_object ):
    keys = [ False, False, False, False ]
    try:
        keys[ 0 ] = neokey_object[ 0 ]
        keys[ 1 ] = neokey_object[ 1 ]
        keys[ 2 ] = neokey_object[ 2 ]
        keys[ 3 ] = neokey_object[ 3 ]
    except ValueError:
        keys = [ False, False, False, False ]
    return keys


def initialize_four_digit_display( i2c_bus ):
    try:
        display = segments.Seg7x4( i2c_bus )
        display.fill(0)
    except ValueError:
        print( "4 digit 7 segment display not found" )
        display = False
    return display

def initialize_neokey1x4( i2c_bus ):
    try:
        neokey = NeoKey1x4(i2c_bus, addr=0x30)
    except ValueError:
        print( "neokey 1x4 module not found" )
        neokey = False
    return neokey

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def initialize_real_time_clock( i2c_bus ):
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        clock_battery_OK = not rtc.battery_low
        if clock_battery_OK:
            print( "clock battery is OK." )
        else:
            print( "clock battery is low. Replace the clock battery." )
    except NameError:
        print( "real time clock failed to initialize" )
        rtc = False
    return ( rtc, clock_battery_OK )

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
    except ValueError as err:
        print( "i2c bus failed to initialize '{}'".format( err ))
        i2c_bus = False
    return i2c_bus


def set_neopixel( neopixel_object, color ):
    if neopixel_object:
        if color:
            neopixel_object[ 0 ] = color
        else:
            neopixel_object[ 0 ] = ( 0, 0, 0 )
        neopixel_object.show()
    else:
        return False

def initialize_neopixel( pin_number, number_of_pixels ):
    try:
        neopixel_object = neopixel.NeoPixel( pin_number, number_of_pixels, brightness=1, auto_write=False, pixel_order=neopixel.GRB )
    except ValueError:
        print( "neopixel initialization failed." )
        neopixel_object = False
    return neopixel_object

def check_battery_voltage( battery_monitor ):
    if battery_monitor:
        battery_voltage = round(( battery_monitor.value * 3.3) / 65536 * 2, 2 )
        return battery_voltage
    else:
        return 0

def initialize_AnalogIn( pin_number ):
    try:
        analog_signal = AnalogIn( pin_number )
    except ValueError:
        print( "analog input initialization failed." )
        analog_signal = False
    return analog_signal

def initialize_discrete_LED( pin_number ):
    try:
        discrete_LED = digitalio.DigitalInOut( pin_number )
        discrete_LED.direction = digitalio.Direction.OUTPUT
    except ValueError:
        print( "discrete LED initialization failed." )
        discrete_LED = False
    return discrete_LED

def stall():
    print("intentionally stalled, press return to continue")
    input_string = False
    while input_string == False:
        input_string = input().strip()

main()
