SOFTWARE_VERSION_NUMBER = "2.0.2"
YEAR_CONSTANT = 2023
# STELLA-Q software
# Science and Technology Education for Land/ Life Assessment
# Paul Mirel

import gc #garbage collection, RAM management
gc.collect()
start_memory_free = gc.mem_free()
print("start memory free {:.1f} kB".format( start_memory_free/1000 ))

import time
start_time = time.monotonic()
# load system libraries
import os
import board
import microcontroller
import digitalio
from analogio import AnalogIn
import neopixel
import sdioio
import storage
import busio
import terminalio
sys_lib_time = time.monotonic()
#print( "time needed to load system libraries == {:.1f} s".format( sys_lib_time - start_time ))

from stella_q_utility import set_date_time

# load device libraries
import adafruit_pcf8523                             # real time clock
import adafruit_sht4x                               # air temp and humidity
import adafruit_tca9548a                            # i2c mux
from adafruit_as726x import AS726x_I2C              # spectral sensors
from adafruit_neokey.neokey1x4 import NeoKey1x4     # neokeyboard
from adafruit_ht16k33 import segments               # 7 segment display driver
import adafruit_mlx90614                            # thermal infrared optional
device_lib_time = time.monotonic()
print( "time needed to load all libraries == {:.1f} s".format( device_lib_time - start_time ))
print( "memory free after load libraries {:.1f} kB".format( gc.mem_free()/1000 ))
gc.collect()
print( "memory free after garbage collection {:.1f} kB".format( gc.mem_free()/1000 ))

VIS_BANDS =         ( 450, 500, 550, 570, 600, 650 ) # from amd as7262 datasheet
VIS_BAND_PREFIXES = ( "V", "B",  "G", "Y", "O", "R" )
NIR_BANDS =         ( 610, 680, 730, 760, 810, 860 ) # from amd as7263 datasheet
AH_ON =     True       #   for active high discrete devices
AH_OFF =    False      #   for active high discrete devices
AL_ON =     False      #   for active low discrete devices
AL_OFF =    True       #   for active low discrete devices
NEO_RED =       ( 255, 0, 0 )
NEO_OFF =       ( 0, 0, 0 )

def main():
    LOW_BATTERY_VOLTAGE = 3.1

    DATA_FILE = "/sd/data.csv"

    RECORD =    0
    HOLD =      1
    LIVE =      2
    AVERAGE =   3

    SHOW_UID =              0
    SHOW_TIME =             1
    SHOW_BATCH_NUMBER =     2
    SHOW_SOFTWARE_VERSION = 3


    HEX_OFF =       0x000000
    HEX_ORANGE =    0xFF5500
    HEX_YELLOW =    0xFFFF00
    HEX_GREEN =     0x00FF00
    HEX_BLUE =      0x0000FF

    MUX_channel_NIR = 0
    MUX_channel_VIS = 1
    power_indicator_LED = initialize_discrete_LED( board.LED )
    power_indicator_LED.value = AH_ON
    operational = True
    # initialize bus
    i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )
    if not i2c_bus:
        operational = False

    if operational:
        # initialize board devices
        UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
        print( "unique identifier (UID) == {0}".format( UID ))
        print( "STELLA software version == {}".format( SOFTWARE_VERSION_NUMBER ))
        main_battery_monitor = initialize_AnalogIn( board.VOLTAGE_MONITOR )
        main_battery_voltage = check_battery_voltage( main_battery_monitor )
        #print( "main_battery_voltage == {}".format( main_battery_voltage ))
        onboard_neopixel = initialize_neopixel( board.NEOPIXEL, 1 )
        set_neopixel( onboard_neopixel, NEO_RED )
        time.sleep( 0.1 )
        set_neopixel( onboard_neopixel, NEO_OFF )
        gc.collect()
        print( "memory free after initialize board devices {:.1f} kB".format( gc.mem_free()/1000 ))

    if operational:
        # initialize hardware devices
        rtc_group = initialize_real_time_clock( i2c_bus )
        rtc = rtc_group[ 0 ]
        rtc_battery_OK = rtc_group[ 1 ]
        days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        # wday =    0         1         2           3            4          5          6
        if rtc:
            timestamp = rtc.datetime
            if timestamp.tm_year < 2023:
                timestamp = time.struct_time(( 2023,  01,   01,   12,  00,  00,   0,   -1,    -1 ))
                rtc.datetime = timestamp
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))

    # initialize sd card file storage
    try:
        sdcard = sdioio.SDCard( clock=board.SDIO_CLOCK, command=board.SDIO_COMMAND, data=board.SDIO_DATA, frequency=25000000 )
        vfs = storage.VfsFat(sdcard)
    except( OSError, ValueError ):
        print( "No SD card found, or card is full" )
        vfs = False
    if vfs:
        storage.mount(vfs, "/sd")
    else:
        set_neopixel( onboard_neopixel, NEO_RED )
        print( "(future print to instrument display): SD card initialization failed. No SD card or card full." )
    write_header = initialize_data_file( DATA_FILE )
    #print( "write header == {}".format( write_header ))
    gc.collect()
    print( "memory free after initialize virtual file system {:.1f} kB".format( gc.mem_free()/1000 ))

    if operational:
        batch_number = update_batch( timestamp )
        ( desired_sample_interval_s, averaging_number_of_samples,
            averaging_count_down, count_down_step_s ) = check_STELLA_config()

        # initialize sensors
        TIR_sensor = initialize_TIR_sensor( i2c_bus )
        temp_hum_SHT40 = initialize_SHT40( i2c_bus )
        try:
            bus_mux = initialize_tca_i2c_multiplexer( i2c_bus )
            NIR_spectral_sensor = initialize_spectral_sensor( bus_mux, MUX_channel_NIR )
            VIS_spectral_sensor = initialize_spectral_sensor( bus_mux, MUX_channel_VIS )
        except OSError:
            bus_mux = False
            NIR_spectral_sensor = False
            VIS_spectral_sensor = False
            print( "spectral sensors not found" )
        gc.collect()
        print( "memory free after initialize sensors {:.1f} kB".format( gc.mem_free()/1000 ))

        # set instrument mode
        instrument_mode = LIVE
        last_instrument_mode = HOLD
        four_digit_display = initialize_four_digit_display( i2c_bus )
        if four_digit_display:
            four_digit_display.print( UID )
        neokey1x4 = initialize_neokey1x4( i2c_bus )
        if neokey1x4:
            neokey1x4.pixels[0] = HEX_ORANGE    # orange key changes display modes
            neokey1x4.pixels[1] = HEX_YELLOW    # yellow key takes one sample, or toggles set_check in date/time mode
            neokey1x4.pixels[2] = HEX_GREEN     # green key takes multiple samples and averages them, or adds to value in date/time mode
            neokey1x4.pixels[3] = HEX_BLUE      # blue toggles between continuous record and pause, or subtracts from value in date/time mo
        gc.collect()
        print( "memory free after initialize keys and display {:.1f} kB".format( gc.mem_free()/1000 ))
    # set modes and controls
    display_mode = SHOW_UID
    number_of_display_modes = 4
    last_display_mode = 10 # startup
    orange_key_pressed =        False
    last_orange_key_pressed =   False
    blue_key_pressed =          False
    last_blue_key_pressed =     False
    yellow_key_pressed =        False
    last_yellow_key_pressed =   False
    green_key_pressed =         False
    last_green_key_pressed =    False

    while operational:
        gc.collect()
        last_sample_time_s = time.monotonic()
        if rtc:
            timestamp = rtc.datetime
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
        iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
        weekday = timestamp_to_weekday( timestamp )
        decimal_hour = timestamp_to_decimal_hour( timestamp )
        print( "\n{} kB memory free at top of the loop at {}\n".format( gc.mem_free()/1000, iso8601_utc ))

        if four_digit_display:
            if display_mode == SHOW_UID and last_display_mode != SHOW_UID:
                four_digit_display.fill(0)
                four_digit_display.print( UID )
                last_display_mode = SHOW_UID
            elif display_mode == SHOW_TIME:
                four_digit_display.fill(0)
                four_digit_display.print( "{:02}:{:02}".format( timestamp.tm_hour, timestamp.tm_min ))
                last_display_mode = SHOW_TIME
            elif display_mode == SHOW_BATCH_NUMBER:
                four_digit_display.fill(0)
                four_digit_display.print( batch_number )
                four_digit_display[0] = ( "B" )
                last_display_mode = SHOW_BATCH_NUMBER
            elif display_mode == SHOW_SOFTWARE_VERSION:
                four_digit_display.fill(0)
                four_digit_display.print( SOFTWARE_VERSION_NUMBER )
                four_digit_display.set_digit_raw(0, 0x1C)
                last_display_mode = SHOW_SOFTWARE_VERSION

        exit_wait = 0

        if instrument_mode == RECORD:
            #print( "instrument mode == RECORD" )
            if last_instrument_mode != RECORD:
                batch_number = update_batch( timestamp )
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
            if VIS_spectral_sensor:
                visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
            else:
                visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            if NIR_spectral_sensor:
                nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
            else:
                nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            surface_temperature_C = read_TIR_sensor( TIR_sensor )
            air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
            main_battery_voltage = check_battery_voltage( main_battery_monitor )
            datapoint = build_datapoint(
                UID, timestamp, decimal_hour, batch_number,
                visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
                surface_temperature_C, air_temperature_C, relative_humidity_percent, main_battery_voltage, SOFTWARE_VERSION_NUMBER
                )
            write_line_success = write_line_to_file( onboard_neopixel, datapoint, DATA_FILE )
            mirror_data_over_usb(
                UID, timestamp, decimal_hour, batch_number,
                visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
                surface_temperature_C, air_temperature_C, relative_humidity_percent)
            last_instrument_mode = RECORD

        if instrument_mode == HOLD:
            #print( "instrument mode == HOLD" )
            if last_instrument_mode != HOLD:
                batch_number = update_batch( timestamp )
                exit_wait = 1
                display_mode = SHOW_BATCH_NUMBER
                if VIS_spectral_sensor:
                    visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                else:
                    visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                if NIR_spectral_sensor:
                    nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                else:
                    nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                surface_temperature_C = read_TIR_sensor( TIR_sensor )
                air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                datapoint = build_datapoint(
                    UID, timestamp, decimal_hour, batch_number,
                    visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent, main_battery_voltage, SOFTWARE_VERSION_NUMBER
                    )
                write_line_success = write_line_to_file( onboard_neopixel, datapoint, DATA_FILE )
                mirror_data_over_usb(
                    UID, timestamp, decimal_hour, batch_number,
                    visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent)
            last_instrument_mode = HOLD

        if instrument_mode == LIVE:
            #print( "instrument mode == LIVE" )
            if VIS_spectral_sensor:
                visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
            else:
                visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            if NIR_spectral_sensor:
                nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
            else:
                nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
            surface_temperature_C = read_TIR_sensor( TIR_sensor )
            air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
            main_battery_voltage = check_battery_voltage( main_battery_monitor )
            mirror_data_over_usb(
                UID, timestamp, decimal_hour, batch_number,
                visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
                surface_temperature_C, air_temperature_C, relative_humidity_percent)
            last_instrument_mode = LIVE

        if instrument_mode == AVERAGE and four_digit_display:
            if vfs:
                if neokey1x4:
                    neokey1x4.pixels[2] = 0x00FF00
                four_digit_display.fill( 0 )
                #print( "instrument mode == AVERAGE" )
                batch_number = update_batch( timestamp )
                exit_wait = 1
                count_down_index = averaging_count_down
                surface_temperature_C = read_TIR_sensor( TIR_sensor )
                air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                VIS_sensor_temperature, NIR_sensor_temperature = get_spectral_sensor_temperatures( VIS_spectral_sensor, NIR_spectral_sensor )
                header_values = ( UID, batch_number, iso8601_utc, surface_temperature_C, air_temperature_C, relative_humidity_percent, VIS_sensor_temperature, NIR_sensor_temperature )
                sampling_filename = create_sampling_file( UID, batch_number, timestamp, decimal_hour, header_values )
                while count_down_index > 0:
                    print( "count down remaining = {}".format( count_down_index ))
                    four_digit_display.fill( 0 )
                    four_digit_display[1] = "C"
                    four_digit_display[2] ="D"
                    four_digit_display.print( count_down_index )
                    count_down_index = count_down_index - 1
                    time.sleep( count_down_step_s )
                sampling_index = averaging_number_of_samples
                sample_average = [ 0,0,0,0,0,0,0,0,0,0,0,0 ]
                while sampling_index > 0:
                    print( "sample number = {}".format( sampling_index ))
                    if VIS_spectral_sensor:
                        visible_spectrum_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                    else:
                        visible_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                    if NIR_spectral_sensor:
                        nir_spectrum_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                    else:
                        nir_spectrum_uW_per_cm_squared = (0,0,0,0,0,0)
                    write_sampling_line_to_file( UID, batch_number, iso8601_utc, onboard_neopixel, sampling_filename, sampling_index, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared )
                    spectral_samples = visible_spectrum_uW_per_cm_squared + nir_spectrum_uW_per_cm_squared
                    #print( spectral_samples )
                    for i,v in enumerate( spectral_samples ):
                        sample_average[ i ] = sample_average[ i ] + v
                    four_digit_display.fill( 0 )
                    four_digit_display.print( sampling_index )
                    sampling_index = sampling_index - 1
                for i,v in enumerate( sample_average ):
                    sample_average[ i ] = round( sample_average[ i ]/ averaging_number_of_samples, 1 )
                #print( sample_average )
                write_tail_of_sampling_file( UID, batch_number, iso8601_utc,
                    onboard_neopixel, sampling_filename, averaging_number_of_samples,
                    sample_average, VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
                    air_temperature_C, relative_humidity_percent
                    )
                four_digit_display.print( "DONE" )
                averaged_visible_spectrum_uW_per_cm_squared = [0,0,0,0,0,0]
                averaged_nir_spectrum_uW_per_cm_squared = [0,0,0,0,0,0]
                for i,v in enumerate( sample_average ):
                    if i < 6:
                        averaged_visible_spectrum_uW_per_cm_squared[ i ] = sample_average[ i ]
                    else:
                        averaged_nir_spectrum_uW_per_cm_squared[ i - 6 ] = sample_average[ i ]
                datapoint = build_datapoint(
                    UID, timestamp, decimal_hour, batch_number,
                    averaged_visible_spectrum_uW_per_cm_squared, averaged_nir_spectrum_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent, main_battery_voltage, SOFTWARE_VERSION_NUMBER)
                write_line_success = write_line_to_file( onboard_neopixel, datapoint, DATA_FILE )
                gc.collect()
                mirror_data_over_usb(
                    UID, timestamp, decimal_hour, batch_number,
                    averaged_visible_spectrum_uW_per_cm_squared, averaged_nir_spectrum_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent)
                time.sleep( 1 )
                display_mode = SHOW_BATCH_NUMBER
                last_instrument_mode = HOLD
                instrument_mode = HOLD
            else:
                print( "print 'no sd' to display" )

        # Wait loop
        gc.collect()
        wait_memory_free = gc.mem_free()
        if main_battery_voltage < LOW_BATTERY_VOLTAGE:
            low_battery_voltage_notification( four_digit_display )
        #print("at wait loop, memory free {} B".format( wait_memory_free ))
        remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
        while remaining_wait_time > 0 and exit_wait == 0:
            # check for keys pressed
            orange_key_pressed, yellow_key_pressed, green_key_pressed, blue_key_pressed = check_neokey( neokey1x4 )
            # turn key light off to acknowledge press detected
            if neokey1x4:
                if orange_key_pressed:
                    neokey1x4.pixels[0] = HEX_OFF
                else:
                    neokey1x4.pixels[0] = HEX_ORANGE
                if yellow_key_pressed:
                    neokey1x4.pixels[1] = HEX_OFF
                else:
                    neokey1x4.pixels[1] = HEX_YELLOW
                if green_key_pressed:
                    neokey1x4.pixels[2] = HEX_OFF
                else:
                    neokey1x4.pixels[2] = HEX_GREEN
                if blue_key_pressed:
                    neokey1x4.pixels[3] = HEX_OFF
                else:
                    neokey1x4.pixels[3] = HEX_BLUE


            if orange_key_pressed and not last_orange_key_pressed:
                print( "check to see if orange key is being held down for a full 2 seconds" )
                loop_count = 0
                while check_neokey( neokey1x4 )[ 0 ] and loop_count < 30:
                    loop_count += 1
                    time.sleep( 0.01 )
                #print( "loop count = {}". format ( loop_count ))
                if loop_count > 25:
                    display_mode = SHOW_BATCH_NUMBER
                    print( "go into date and time set mode" ) # ------------ date and time set mode -------------------------------------
                    neokey1x4.pixels[0] = 0xFF5500
                    gc.collect()
                    print( "memory free before going into date time update = {}".format( gc.mem_free() ))
                    updated_timestamp = set_date_time( rtc, timestamp, four_digit_display, neokey1x4 ) #check and update the date and time
                    print( "update_timestamp == {}".format( updated_timestamp ))
                    try:
                        rtc.datetime = updated_timestamp
                    except ValueError:
                        print( "Error: unable to update hardware real time clock" )
                    while neokey1x4[ 0 ]:
                        print( "waiting for orange key release" )
                        time.sleep( 0.01 )
                    while not neokey1x4[ 0 ]:
                        timestamp = rtc.datetime
                        four_digit_display.print("{:02}:{:02}".format( timestamp.tm_min, timestamp.tm_sec ))
                        print( "waiting for orange key press" )
                        time.sleep( 0.25 )
                display_mode = ( display_mode + 1 ) % number_of_display_modes
                if instrument_mode == HOLD:
                    instrument_mode = LIVE
            last_orange_key_pressed = orange_key_pressed

            if yellow_key_pressed and not last_yellow_key_pressed:
                if instrument_mode != HOLD:
                    instrument_mode = HOLD
                last_instrument_mode = RECORD
            last_yellow_key_pressed = yellow_key_pressed

            if blue_key_pressed and not last_blue_key_pressed:
                if instrument_mode != RECORD:
                    instrument_mode = RECORD
                else:
                    instrument_mode = LIVE
            last_blue_key_pressed = blue_key_pressed

            if green_key_pressed and not last_green_key_pressed:
                if instrument_mode != AVERAGE:
                    instrument_mode = AVERAGE
            last_green_key_pressed = green_key_pressed

            #print( display_mode )
            remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
            #print( "remaining_wait_time == {:}".format(remaining_wait_time))
            time.sleep( 0.1 )


    print( "STELLA done" )

# define functions

gc.collect()

def mirror_data_over_usb(
    UID, timestamp, decimal_hour, batch_number,
    visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
    surface_temperature_C, air_temperature_C, relative_humidity_percent
    ):
    gc.collect()
    mirror_data = (
        "mirror_data, UID, {}, batch, {}, year, {}, month, {}, "
        "day, {}, hour, {}, min, {}, sec, {}, dec_hour, {}, "
        "v450, {}, b500, {}, g550, {}, y570, {}, o600, {}, r650, {}, "
        "610, {}, 680, {}, 730, {}, 760, {}, 810, {}, 860, {}, "
        "surface_temp, {}, air_temp, {}, rel_humidity, {}"
        .format( str(UID), str(batch_number), str( timestamp.tm_year), str( timestamp.tm_mon ),
        str( timestamp.tm_mday ), str( timestamp.tm_hour ), str( timestamp.tm_min ), str( timestamp.tm_sec ), str( decimal_hour ),
        str( visible_spectrum_uW_per_cm_squared[ 0 ] ),
        str( visible_spectrum_uW_per_cm_squared[ 1 ] ),
        str( visible_spectrum_uW_per_cm_squared[ 2 ] ),
        str( visible_spectrum_uW_per_cm_squared[ 3 ] ),
        str( visible_spectrum_uW_per_cm_squared[ 4 ] ),
        str( visible_spectrum_uW_per_cm_squared[ 5 ] ),
        str( nir_spectrum_uW_per_cm_squared[ 0 ]),
        str( nir_spectrum_uW_per_cm_squared[ 1 ]),
        str( nir_spectrum_uW_per_cm_squared[ 2 ]),
        str( nir_spectrum_uW_per_cm_squared[ 3 ]),
        str( nir_spectrum_uW_per_cm_squared[ 4 ]),
        str( nir_spectrum_uW_per_cm_squared[ 5 ]),
        str( surface_temperature_C ), str( air_temperature_C ), str( relative_humidity_percent )
        ))

    print( mirror_data )
    return True

def check_neokey( neokey_object ):
    gc.collect()
    keys = [ False, False, False, False ]
    if neokey_object:
        try:
            keys[ 0 ] = neokey_object[ 0 ]
            keys[ 1 ] = neokey_object[ 1 ]
            keys[ 2 ] = neokey_object[ 2 ]
            keys[ 3 ] = neokey_object[ 3 ]
        except ValueError:
            keys = [ False, False, False, False ]
    return keys

def read_TIR_sensor( TIR_sensor ):
    if TIR_sensor:
        decimal_places = 1
        try:
            surface_temperature_C = TIR_sensor.object_temperature
            surface_temperature_C = round( surface_temperature_C, decimal_places )
            return surface_temperature_C
        except ValueError as err:
            print( "Error: thermal infrared sensor fail: {:}".format(err) )
            return -273
    else:
        return -273


def read_environmental_sensors( temp_hum_SHT40 ):
    gc.collect()
    if temp_hum_SHT40:
        air_temperature_C, relative_humidity_percent = temp_hum_SHT40.measurements
        air_temperature_C = round( air_temperature_C, 1 )
        relative_humidity_percent = round( relative_humidity_percent, 1 )
    else:
        air_temperature_C = 0
        relative_humidity_percent = 0
    return air_temperature_C, relative_humidity_percent

def read_spectrum( spectral_sensor ):
    gc.collect()
    decimal_places = 1
    # returns readings in a tuple, six channels of data
    # Calibration information from AS7262 datasheet:
    # Each channel is tested with GAIN = 16x, Integration Time (INT_T) = 166ms and VDD = VDD1 = VDD2 = 3.3V, TAMB=25°C.
    # The accuracy of the channel counts/μW/cm2 is ±12%.
    # Sensor Mode 2 is a continuous conversion of light into data on all 6 channels
    # 450nm, 500nm, 550nm, 570nm, 600nm and 650nm
    # sensor.violet returns the calibrated floating point value in the violet channel.
    # sensor.raw_violet returns the uncalibrated decimal count value in the violet channel.
    # that syntax is the same for each of the 6 channels

    # NOTE: the library was written for the visible sensor, as7262.
    # The library works exactly the same for the near infrared sensor, as7263,
    # but the band names are called as if the sensor were the visible one.
    try:
        while not spectral_sensor.data_ready:
            time.sleep(0.01)
        initial_read_time_s = time.monotonic()
        violet_calibrated = round( spectral_sensor.violet, decimal_places )
        blue_calibrated = round( spectral_sensor.blue, decimal_places )
        green_calibrated = round( spectral_sensor.green, decimal_places )
        yellow_calibrated = round( spectral_sensor.yellow, decimal_places )
        orange_calibrated = round( spectral_sensor.orange, decimal_places )
        red_calibrated = round( spectral_sensor.red, decimal_places )
        final_read_time_s = time.monotonic()
        read_time_s = final_read_time_s - initial_read_time_s
        #print( "Sensor read time s = {}".format( read_time_s ))
        spectral_tuple = ( violet_calibrated, blue_calibrated, green_calibrated, yellow_calibrated, orange_calibrated, red_calibrated )
        return spectral_tuple
    except ValueError as err:
        print( "Error: spectral sensor fail: {:}".format(err) )
        return ( 0, 0, 0, 0, 0, 0)

def low_battery_voltage_notification( four_digit_display ):
    four_digit_display.print( "LO B" )
    print( "main battery low, plug in to recharge" )

def write_tail_of_sampling_file( UID, batch_number, iso8601_utc,
    onboard_neopixel, sampling_filename, averaging_number_of_samples,
    sample_average, VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
    air_temperature_C, relative_humidity_percent
    ):
    gc.collect()
    if onboard_neopixel:
        set_neopixel( onboard_neopixel, NEO_RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( batch_number ))
            sf.write( ", average, " )
            for i,v in enumerate( sample_average ):
                sf.write( str(v) )
                #if i < len(sample_average)-1: # no trailing ','
                sf.write(",")
            #sf.write( "\n" )
            sf.write( " {}, {}, {}, {}, {}, {}, {}\n".format(
                VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
                air_temperature_C, relative_humidity_percent, UID, iso8601_utc
                ))
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def write_sampling_line_to_file( UID, batch_number, iso8601_utc, onboard_neopixel, sampling_filename, sampling_index, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared ):
    gc.collect()
    with open( sampling_filename, "r" ) as sf:
        sf.flush()
        sf.close()

    if onboard_neopixel:
        set_neopixel( onboard_neopixel, NEO_RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( batch_number ))
            sf.write( ", " )
            sf.write( str( sampling_index ))
            sf.write( ", " )
            for i,v in enumerate(visible_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                sf.write(",")
            for i,v in enumerate(nir_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                #if i < len(nir_spectrum_uW_per_cm_squared)-1: # no trailing ','
                sf.write(",")
            sf.write( "0, 0, 0, 0, 0, ")
            sf.write( str( UID ))
            sf.write(",")
            sf.write( str( iso8601_utc ))
            sf.write( "\n" )
            sf.flush()
            sf.close()
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def get_spectral_sensor_temperatures(VIS_sensor, NIR_sensor):
    vis_temp_C = round( VIS_sensor.temperature, 1 )
    nir_temp_C = round( NIR_sensor.temperature, 1 )
    return vis_temp_C, nir_temp_C

def create_sampling_file( UID, batch_number, timestamp, decimal_hour, values ):
    gc.collect()
    sampling_filename = "/sd/{}_{:04}{:02}{:02}_{}.csv".format( UID, timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday, batch_number)
    try:
        with open( sampling_filename, "w" ) as sf:
            sf.write( "batch, sample_number, V450_irradiance, B500_irradiance, G550_irradiance, Y570_irradiance, O600_irradiance, "
                "R650_irradiance, nir610_irradiance, nir680_irradiance, nir730_irradiance, nir760_irradiance, "
                "nir810_irradiance, nir860_irradiance,  VIS_sensor_temperature,  NIR_sensor_temperature, "
                "surface_temperature, air_temperature, relative_humidity, UID, timestamp\n" )
            sf.close()
        return sampling_filename
    except OSError:
        return False

def write_line_to_file( onboard_neopixel, values, DATA_FILE ):
    gc.collect()
    try:
        # with sys.stdout as f: # uncomment to write to console (comment "with open")
        file_handling_last_sample_time_s = time.monotonic()
        with open( DATA_FILE, "a" ) as f: # open seems about TBD secs.
            last_sample_time_s = time.monotonic()
            if onboard_neopixel:
                set_neopixel( onboard_neopixel, NEO_RED )
            for i,v in enumerate(values): # 0.16 secs to write
                f.write( str(v) )
                #print( str(v) )
                if i < len(values)-1: # no trailing ','
                    f.write(",")
            f.write("\n")
        f.close()
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        write_command_time = time.monotonic() - last_sample_time_s
        #print ( "write command time s = {}".format(write_command_time))
        file_write_time = time.monotonic() - file_handling_last_sample_time_s
        print ( "File write time s = {}".format(file_write_time))
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "Error: sd card fail: {:} ".format(err) )
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_RED ) # constant ON to show error, likely no SD card present, or SD card full.
        return False
    return True

def build_datapoint(
    UID, timestamp, decimal_hour, batch_number,
    visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared, surface_temperature_C,
    air_temperature_C, relative_humidity_percent, main_battery_voltage, SOFTWARE_VERSION_NUMBER):
    gc.collect()
    iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
    weekday = timestamp_to_weekday( timestamp )
    decimal_hour = timestamp_to_decimal_hour( timestamp )
    datapoint = [] #reset datapoint string
    #datapoint.append( "UID:" )
    datapoint.append( UID )
    datapoint.append( batch_number )
    datapoint.append( "{:},{:},hh.hhhh,{:}".format(weekday, iso8601_utc, decimal_hour ) )
    # name,units,value,calibration_error
    datapoint.append( "surface_temperature,C,{:},1.0".format( surface_temperature_C ) )
    datapoint.append( "air_temperature,C,{:},0.3".format( air_temperature_C ) )
    datapoint.append( "relative_humidity,%,{:},1.8".format(  relative_humidity_percent ) )
    # spectrum, wavelength-units, irradiance-units, bandlabel, irradiance, cal-error
    datapoint.append( "visible_spectrum,nm,uW/cm^2,12/100")
    # VIS_BANDS have a color initial prefix
    for i,band in enumerate( VIS_BANDS ):
        datapoint.append( band )
        datapoint.append( visible_spectrum_uW_per_cm_squared[ i ] )
    datapoint.append( "near_infrared_spectrum,nm,uW/cm^2,12/100")
    for i,band in enumerate( NIR_BANDS ):
        datapoint.append( band )
        datapoint.append( nir_spectrum_uW_per_cm_squared[ i ] )
    datapoint.append( "battery_voltage,V,{:}".format( main_battery_voltage ) )
    datapoint.append( "{}".format( SOFTWARE_VERSION_NUMBER ) )
    return datapoint

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def initialize_four_digit_display( i2c_bus ):
    try:
        display = segments.Seg7x4( i2c_bus )
        print("display succeeded")
        display.fill(0)
    except ValueError:
        print( "4 digit 7 segment display not found" )
        display = False
    return display


def initialize_neokey1x4( i2c_bus ):
    try:
        neokey = NeoKey1x4(i2c_bus, addr=0x30)
    except ValueError:
        print( "neokey 1x4 module not found" )
        neokey = False
    return neokey

def initialize_spectral_sensor( bus_mux, mux_channel ):
    try:
        spectral_sensor = AS726x_I2C( bus_mux[ mux_channel ])
        #print( "\nSet spectral sensor at mux channel {} to 16x gain and 166 ms integration time, because those are the values used at the factory during the sensor calibration.".format( mux_channel ))
        spectral_sensor.conversion_mode = spectral_sensor.MODE_2
        #print( "Default Gain = {}".format( spectral_sensor.gain ))
        spectral_sensor.gain = 16
        #print( "MUX channel {}: Gain = {}".format( mux_channel, spectral_sensor.gain ))
        #print( "Default Integration Time = {} ms".format( spectral_sensor.integration_time ))
        spectral_sensor.integration_time = 166 # from 1 to 714 = 255 * 2.8ms ms
        print( "Spectral Sensor on MUX channel {}: Gain = {}x, Integration Time = {} ms".format( mux_channel, spectral_sensor.gain, spectral_sensor.integration_time ))
        return spectral_sensor
    except ValueError:
        print( "spectral_sensor at channel {} not found".format( mux_channel ))
        return False

def initialize_tca_i2c_multiplexer( i2c_bus ):
    try:
        tca = adafruit_tca9548a.TCA9548A( i2c_bus )
    except ValueError:
        print( "i2c bus multiplexer not available" )
        tca = False
    return tca

def initialize_TIR_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        TIR_sensor = adafruit_mlx90614.MLX90614( i2c_bus )
        print( "MLX90614 remote surface thermometer initialized" )
        return TIR_sensor
    except ( ValueError, NameError) as err:
        print( "Error: thermal infrared sensor fail: {:}".format(err) )
        return False

def initialize_SHT40( i2c_bus ):
    try:
        sht = adafruit_sht4x.SHT4x( i2c_bus )
    except ValueError:
        print( "sht40 temp_humid could not be found" )
        sht = False
    try:
        #print( "Found SHT40 with serial number", hex( sht.serial_number ))
        sht.mode = adafruit_sht4x.Mode.NOHEAT_HIGHPRECISION
        # Can also set the mode to enable heater
        # sht.mode = adafruit_sht4x.Mode.LOWHEAT_100MS
        print("SHT40 air temperature and humidity mode: {}".format( adafruit_sht4x.Mode.string[ sht.mode ]))
    except( OSError, AttributeError ):
        print( "SHT40 sensor serial number lookup failed" )
    return sht

def check_STELLA_config():
    try:
        with open( "/sd/STELLA_config.txt", "r" ) as s:
            try:
                header = s.readline()
                #print( sampling_header )
                values = s.readline()
                #print(sampling_values)
                values= values.rstrip('\n')
                values = values.split(',')
                #print(line)
                desired_sample_interval_s = float( values[ 0 ])
                averaging_number_of_samples = int( values[ 1 ])
                averaging_count_down = int( values[ 2 ])
                count_down_step_s = float( values[ 3 ])

            # TBD: catch error when /sd doesn't exist
            except ValueError:
                pass
    except OSError:
            print( "STELLA_config.txt file not found" )
            try:
                with open( "/sd/STELLA_config.txt", "w" ) as s:
                    s.write( "desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s\n" )
                    s.write( "0.75, 20, 3, 0.75\n" )
                print("created new STELLA_config.txt file")
            except OSError:
                pass
            desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s = 0.75, 21, 3, 0.75
    return desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s

def update_batch( timestamp ):
    gc.collect()
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            # TBD: catch error when /sd doesn't exist
            except ValueError:
                previous_batch_number = 0
                # corrupted data in batch number file, setting batch to 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0
    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
        # TBD: catch error when /sd doesn't exist
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def initialize_data_file( DATA_FILE ):
    need_header = False
    try:
        os.stat( DATA_FILE ) # fail if data.csv file is not already there
        #raise OSError # uncomment to force header everytime
        return False
    except OSError:
        # setup the header for first time
        need_header = True
        #print( "need header = True")
        if need_header:
            preheader_mem_free = gc.mem_free()
            print( "preheader mem free = {}".format(preheader_mem_free))
            gc.collect()
            try:
                with open( DATA_FILE, "w" ) as f:
                    header = (
                        "UID, batch_number, weekday, timestamp, hour_format, decimal_hour, "
                        "surface_temperature_name, surface_temperature_units, surface_temperature, surface_temperature_calibration, "
                        "air_temperature_name, air_temperature_units, air_temperature, air_temperature_calibration, "
                        "relative_humidity_name, relative_humidity_units, relative_humidity, relative_humidity_calibration, "
                        "visible_spectrum_name, visible_spectrum_wavelength, visible_spectrum_units, visible_spectrum_calibration, "
                        "V450_wavelength, V450_irradiance, B500_wavelength, B500_irradiance, G550_wavelength, G550_irradiance, "
                        "Y570_wavelength, Y570_irradiance, O600_wavelength, O600_irradiance, R650_wavelength, R650_irradiance, "
                        "near_infrared_spectrum_name, near_infrared_spectrum_wavelength, near_infrared_spectrum_units, near_infrared_spectrum_calibration, "
                        "nir610_wavelength, nir610_irradiance, nir680_wavelength, nir680_irradiance, nir730_wavelength, nir730_irradiance, "
                        "nir760_wavelength, nir760_irradiance, nir810_wavelength, nir810_irradiance, nir860_wavelength, nir860_irradiance, "
                        "battery_voltage_name, battery_voltage_units, battery_voltage, SOFTWARE_VERSION_NUMBER"
                        "\n"
                        )

                    f.write( header )
            except OSError:
                pass
            postheader_mem_free = gc.mem_free()
            print( "header mem usage = {}".format(postheader_mem_free - preheader_mem_free))
        return True

print( "memory check in the midst of function definitions, after the header definition = {}".format( gc.mem_free()))
gc.collect()
print( "memory check after garbage collection = {}".format( gc.mem_free()))

def initialize_real_time_clock( i2c_bus ):
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        clock_battery_OK = not rtc.battery_low
        if clock_battery_OK:
            print( "clock battery is OK." )
        else:
            print( "clock battery is low. Replace the clock battery." )
    except ( NameError, ValueError ):
        print( "real time clock failed to initialize" )
        rtc = False
        clock_battery_OK = False
    return ( rtc, clock_battery_OK )

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
    except ValueError as err:
        print( "i2c bus failed to initialize '{}'".format( err ))
        i2c_bus = False
    return i2c_bus

def set_neopixel( neopixel_object, color ):
    if neopixel_object:
        if color:
            neopixel_object[ 0 ] = color
        else:
            neopixel_object[ 0 ] = ( 0, 0, 0 )
        neopixel_object.show()
    else:
        return False

def initialize_neopixel( pin_number, number_of_pixels ):
    try:
        neopixel_object = neopixel.NeoPixel( pin_number, number_of_pixels, brightness=1, auto_write=False, pixel_order=neopixel.GRB )
    except ValueError:
        print( "neopixel initialization failed." )
        neopixel_object = False
    return neopixel_object

def check_battery_voltage( battery_monitor ):
    if battery_monitor:
        battery_voltage = round(( battery_monitor.value * 3.3) / 65536 * 2, 2 )
        return battery_voltage
    else:
        return 0

def initialize_discrete_LED( pin_number ):
    try:
        discrete_LED = digitalio.DigitalInOut( pin_number )
        discrete_LED.direction = digitalio.Direction.OUTPUT
    except ValueError:
        print( "discrete LED initialization failed." )
        discrete_LED = False
    return discrete_LED

def initialize_AnalogIn( pin_number ):
    try:
        analog_signal = AnalogIn( pin_number )
    except ValueError:
        print( "analog input initialization failed." )
        analog_signal = False
    return analog_signal

print("memory free after loading function definitions into RAM = {} bytes".format( gc.mem_free() ))

main()
