#set_time_date_utility
YEAR_CONSTANT = 2023
# STELLA-Q utility software
# Science and Technology Education for Land/ Life Assessment
# Paul Mirel

import gc #garbage collection, RAM management
gc.collect()
start_memory_free = gc.mem_free()
print("start memory free {:.1f} kB".format( start_memory_free/1000 ))

import time
start_time = time.monotonic()
# load system libraries
import os
import board
import microcontroller
import digitalio
import busio
import terminalio
sys_lib_time = time.monotonic()
#print( "time needed to load system libraries == {:.1f} s".format( sys_lib_time - start_time ))

# load device libraries
import adafruit_pcf8523                             # real time clock
from adafruit_neokey.neokey1x4 import NeoKey1x4     # neokeyboard
from adafruit_ht16k33 import segments               # 7 segment display driver
device_lib_time = time.monotonic()
print( "time needed to load all libraries == {:.1f} s".format( device_lib_time - start_time ))
print( "memory free after load libraries {:.1f} kB".format( gc.mem_free()/1000 ))
gc.collect()
print( "memory free after garbage collection {:.1f} kB".format( gc.mem_free()/1000 ))

from stella_q_utility import set_date_time

def main():
    number_of_display_modes = 3
    SHOW_UID =              0
    SHOW_MESSAGE =          1
    SHOW_MMSS =             2

    HEX_OFF =       0x000000
    HEX_ORANGE =    0xFF5500
    HEX_YELLOW =    0xFFFF00
    HEX_GREEN =     0x00FF00
    HEX_BLUE =      0x0000FF

    has_been_set = False
    power_indicator_LED = initialize_discrete_LED( board.LED )
    power_indicator_LED.value = True
    operational = True
    i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )
    if not i2c_bus:
        operational = False

    # initialize board devices
    UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
    print( "unique identifier (UID) == {0}".format( UID ))
    gc.collect()
    print( "memory free after initialize board devices {:.1f} kB".format( gc.mem_free()/1000 ))
    # initialize hardware devices
    rtc_group = initialize_real_time_clock( i2c_bus )
    rtc = rtc_group[ 0 ]
    rtc_battery_OK = rtc_group[ 1 ]
    days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
    # wday =    0         1         2           3            4          5          6
    if rtc:
        timestamp = rtc.datetime
        if timestamp.tm_year < 2023:
            timestamp = time.struct_time(( 2023,  01,   01,   12,  00,  00,   0,   -1,    -1 ))
            rtc.datetime = timestamp
    else:
        timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))

    four_digit_display = initialize_four_digit_display( i2c_bus )
    if four_digit_display:
        four_digit_display.print( UID )
    neokey1x4 = initialize_neokey1x4( i2c_bus )
    if neokey1x4:
        neokey1x4.pixels[0] = HEX_ORANGE    # orange key changes display modes
        neokey1x4.pixels[1] = HEX_YELLOW    # yellow key takes one sample, or toggles set_check in date/time mode
        neokey1x4.pixels[2] = HEX_GREEN     # green key takes multiple samples and averages them, or adds to value in date/time mode
        neokey1x4.pixels[3] = HEX_BLUE      # blue toggles between continuous record and pause, or subtracts from value in date/time mo
    gc.collect()
    print( "memory free after initialize keys and display {:.1f} kB".format( gc.mem_free()/1000 ))
    # set modes and controls
    display_mode = SHOW_UID
    last_display_mode = 10 # startup
    orange_key_pressed =        False
    last_orange_key_pressed =   False
    blue_key_pressed =          False
    last_blue_key_pressed =     False
    yellow_key_pressed =        False
    last_yellow_key_pressed =   False
    green_key_pressed =         False
    last_green_key_pressed =    False
    startup = True
    startup_time_s = time.monotonic()
    show_UID_time_s = 3
    while operational:
        gc.collect()
        time_now_s = time.monotonic()
        if time_now_s - startup_time_s > show_UID_time_s and time_now_s - startup_time_s < show_UID_time_s + 0.1 :
            display_mode = ( display_mode + 1 ) % number_of_display_modes
            print( "display mode == {}".format( display_mode ))

        if rtc:
            timestamp = rtc.datetime
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
        iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
        weekday = timestamp_to_weekday( timestamp )
        decimal_hour = timestamp_to_decimal_hour( timestamp )
        #print( "\n{} kB memory free at top of the loop at {}\n".format( gc.mem_free()/1000, iso8601_utc ))
        print( "press and hold orange key to set date and time. current date and time = {} {} {} {}:{}:{}"
            .format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec ))

        if display_mode == SHOW_UID and last_display_mode != SHOW_UID:
            four_digit_display.fill(0)
            four_digit_display.print( UID )
            last_display_mode = SHOW_UID
        elif display_mode == SHOW_MESSAGE and last_display_mode != SHOW_MESSAGE:
            four_digit_display.fill(0)
            #four_digit_display.marquee( "SET TO ", 0.25, False )
            four_digit_display.print( "UTC-" )
            last_display_mode = SHOW_MESSAGE
        elif display_mode == SHOW_MMSS and last_display_mode != SHOW_MMSS:
            #four_digit_display.colon = True
            four_digit_display.print("{:02}:{:02}".format( timestamp.tm_min, timestamp.tm_sec ))

        gc.collect()

        orange_key_pressed, yellow_key_pressed, green_key_pressed, blue_key_pressed = check_neokey( neokey1x4 )
        # turn key light off to acknowledge press detected
        if neokey1x4:
            if orange_key_pressed:
                neokey1x4.pixels[0] = HEX_OFF
            else:
                neokey1x4.pixels[0] = HEX_ORANGE
            if yellow_key_pressed:
                neokey1x4.pixels[1] = HEX_OFF
            else:
                neokey1x4.pixels[1] = HEX_YELLOW
            if green_key_pressed:
                neokey1x4.pixels[2] = HEX_OFF
            else:
                neokey1x4.pixels[2] = HEX_GREEN
            if blue_key_pressed:
                neokey1x4.pixels[3] = HEX_OFF
            else:
                neokey1x4.pixels[3] = HEX_BLUE


        if orange_key_pressed and not last_orange_key_pressed:
            print( "check to see if orange key is being held down for a full 2 seconds" )
            loop_count = 0
            display_mode = ( display_mode + 1 ) % number_of_display_modes
            while check_neokey( neokey1x4 )[ 0 ] and loop_count < 30:
                loop_count += 1
                time.sleep( 0.01 )
                print( "loop count = {}". format ( loop_count ))
            if loop_count > 20:
                print( "go into date and time set mode" ) # ------------ date and time set mode -------------------------------------
                neokey1x4.pixels[0] = 0xFF5500
                gc.collect()
                print( "memory free before going into date time update = {}".format( gc.mem_free() ))
                updated_timestamp = set_date_time( rtc, timestamp, four_digit_display, neokey1x4 ) #check and update the date and time
                display_mode = SHOW_MMSS
                print( "update_timestamp == {}".format( updated_timestamp ))
                try:
                    rtc.datetime = updated_timestamp
                except ValueError:
                    print( "Error: unable to update hardware real time clock" )
                has_been_set = True

        last_orange_key_pressed = orange_key_pressed

        time.sleep( 0.1 )


    print( "STELLA done" )

# define functions

gc.collect()

def check_neokey( neokey_object ):
    gc.collect()
    keys = [ False, False, False, False ]
    if neokey_object:
        try:
            keys[ 0 ] = neokey_object[ 0 ]
            keys[ 1 ] = neokey_object[ 1 ]
            keys[ 2 ] = neokey_object[ 2 ]
            keys[ 3 ] = neokey_object[ 3 ]
        except ValueError:
            keys = [ False, False, False, False ]
    return keys

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def initialize_four_digit_display( i2c_bus ):
    try:
        display = segments.Seg7x4( i2c_bus )
        display.fill(0)
    except ValueError:
        print( "4 digit 7 segment display not found" )
        display = False
    return display

def initialize_neokey1x4( i2c_bus ):
    try:
        neokey = NeoKey1x4(i2c_bus, addr=0x30)
    except ValueError:
        print( "neokey 1x4 module not found" )
        neokey = False
    return neokey

def initialize_real_time_clock( i2c_bus ):
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        clock_battery_OK = not rtc.battery_low
        if clock_battery_OK:
            print( "clock battery is OK." )
        else:
            print( "clock battery is low. Replace the clock battery." )
    except ( NameError, ValueError ):
        print( "real time clock failed to initialize" )
        rtc = False
        clock_battery_OK = False
    return ( rtc, clock_battery_OK )

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
    except ValueError as err:
        print( "i2c bus failed to initialize '{}'".format( err ))
        i2c_bus = False
    return i2c_bus

def initialize_discrete_LED( pin_number ):
    try:
        discrete_LED = digitalio.DigitalInOut( pin_number )
        discrete_LED.direction = digitalio.Direction.OUTPUT
    except ValueError:
        print( "discrete LED initialization failed." )
        discrete_LED = False
    return discrete_LED

print("memory free after loading function definitions into RAM = {} bytes".format( gc.mem_free() ))

main()
